@@ -2,7 +2,7 @@ def check_tb_availability():
     try:
         # torch.utils.tensorboard will fail if `tensorboard` is not available,
         # see their docs for more details: https://pytorch.org/docs/1.8.0/tensorboard.html
-        import tensorboard
+        import tensorboard  # noqa: F401
     except ImportError:
         print('If you want to use tensorboard logging, please `pip install tensorboard`')
         raise
@@ -10,7 +10,7 @@ def check_tb_availability():
 
 def check_wandb_availability():
     try:
-        import wandb
+        import wandb  # noqa: F401
     except ImportError:
         print(
             'If you want to use wandb logging, please `pip install wandb` and follow the instructions at https://docs.wandb.ai/quickstart'
